/*
 * Copyright (c) 1991, 1992 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that: (1) source code distributions
 * retain the above copyright notice and this paragraph in its entirety, (2)
 * distributions including binary code include the above copyright notice and
 * this paragraph in its entirety in the documentation or other materials
 * provided with the distribution, and (3) all advertising materials mentioning
 * features or use of this software display the following acknowledgement:
 * ``This product includes software developed by the University of California,
 * Lawrence Berkeley Laboratory and its contributors.'' Neither the name of
 * the University nor the names of its contributors may be used to endorse
 * or promote products derived from this software without specific prior
 * written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */
#ifndef lint
char copyright[] =
    "@(#) Copyright (c) 1991, 1992 The Regents of the University of California.\nAll rights reserved.\n";
static  char rcsid[] =
    "@(#)$Header: histtrim.c,v 1.16 92/01/26 13:32:33 leres Exp $ (LBL)";
#endif

/*
 * histtrim - trim the history file
 */
#include <stdio.h>
#include <sys/types.h>
#include <sys/file.h>
#include <unistd.h>
#ifndef SYSV
#include <strings.h>
#else
#include <string.h>
#define rindex strrchr
#endif
#include <ctype.h>
#include <errno.h>

#include "dexpire.h"
#include "hash.h"
#include "util.h"
#ifdef EXPIRES_DBZ
#include "dbz.h"
#endif

/* Private data */
static char *dfeedback;
static char *history;
static char *nhistory;
static char *expires;
static char *nexpires;
static char *spool_dir;

static time_t currenttime;	/* current timestamp */
static time_t mintime = 0;	/* keep all history entries newer than this */
static time_t maxtime = 365L*24*60*60;
                                /* disregard expires if longer than this */

/* Public data */
char *prog;

int debug = 0;			/* debugging modes */
int verbose = 0;		/* verbose information to stdout */
int doexpires = 0;		/* update the explicit expires file */
int nflag = 0;                  /* dummy */
int pdesc = -1;                 /* descriptor for new history */

/* Statistics */
int bad = 0;
int newnogroup = 0;		/* number of newly created "no groups" */
int oldnogroup = 0;		/* number of "no groups" omitted */
int nogroup = 0;		/* number of "no groups" in new history */
int nofeedback = 0;		/* number not in the feedback file */
int neverexpired = 0;
int emitted = 0;
int numread = 0;		/* number of old history lines read */
int zombies = 0;                /* dummy */

static int histtrim();
static int artexists();

/* External data */

extern char *optarg;
extern int optind, opterr;

extern int errno;

#ifdef EXPIRES_DBZ
extern void strdcpy();
extern void strtcpy();
#endif

int
main(argc, argv)
	int argc;
	char **argv;
{
	register int op;
	register char *cp;
	register FILE *fin, *fout, *fexp;
	int status;
	char *usage = "usage: %s [-duv] [-m days] [-c days]\n";

	if ((cp = rindex(argv[0], '/')))
		prog = cp + 1;
	else
		prog = argv[0];

	dfeedback=savestr(ctlfile("dfeedback"));
	history=savestr(ctlfile("history"));
	nhistory=savestr(ctlfile("history.n"));
	expires=savestr(ctlfile("expires"));
	nexpires=savestr(ctlfile("expires.new"));
	spool_dir=savestr(fullartfile(NULL));
	/* Process arguments */
	while ((op = getopt(argc, argv, "duvm:c:p:")) != EOF)
		switch (op) {

		case 'd':
			++debug;
			break;

		case 'm':
			mintime = atoi(optarg);
			if (mintime <= 0 || mintime > 365) {
				(void) fprintf(stderr,
				    "%s: \"%s\" invalid argument to -m\n",
				    prog, optarg);
				exit(1);
				/* NOTREACHED */
			}
			/* Convert days to seconds */
			mintime *= 24 * 60 * 60;
			break;

		case 'c':
			maxtime = atoi(optarg);
			if (maxtime <= 0 || maxtime > 365) {
				(void) fprintf(stderr,
				    "%s: \"%s\" invalid argument to -M\n",
				    prog, optarg);
				exit(1);
				/* NOTREACHED */
			}
			/* Convert days to seconds */
			maxtime *= 24 * 60 * 60;
			break;

		case 'u':
			++doexpires;
			break;

		case 'v':
			++verbose;
			break;

		case 'p':
			pdesc = atoi(optarg);
			break; /* XX no check */

		default:
			(void) fprintf(stderr, usage, prog);
			exit(1);
			/* NOTREACHED */
		}

	if (optind != argc) {
		(void) fprintf(stderr, usage, prog);
		exit(1);
		/* NOTREACHED */
	}

	/* Fetch current time (used in various calculations) */
	currenttime = time(0);

	/* Convert from delta to timestamp */
	mintime = currenttime - mintime;
	maxtime = currenttime + maxtime;

	/* Report various times */
	if (verbose) {
		(void) printf("%s: Current time: %s\n",
		    prog, fmtdate(currenttime));
		if (mintime != currenttime)
			(void) printf("%s: Minimum time: %s\n",
			    prog, fmtdate(mintime));
		else
			(void) printf("%s: No minimum time\n", prog);
		(void) fflush(stdout);
	}

	/* Load in the feedback file */
	if ((fin = fopen(dfeedback, "r")) == 0) {
		(void) fprintf(stderr, "%s: fopen(): ", prog);
		perror(dfeedback);
		exit(1);
		/* NOTREACHED */
	}
	if (!readfeedback(fin, hashadd))
		exit(1);
	(void)fclose(fin);

	/* Dump hash table if debugging */
	if (debug > 2)
		hashdump();

	/* Open the current history file */
	if ((fin = fopen(history, "r")) == 0) {
		(void) fprintf(stderr, "%s: fopen(): ", prog);
		perror(history);
		exit(1);
		/* NOTREACHED */
	}

	/* Create the new history file */
	if (pdesc >= 0) {
	    if ((fout = fdopen(pdesc, "w")) == 0) {
		(void) fprintf(stderr, "%s: fopen(): ", prog);
		perror("-p");
		exit(1);
		/* NOTREACHED */
	    }
	} else {
	    if ((fout = fopen(nhistory, "w")) == 0) {
		(void) fprintf(stderr, "%s: fopen(): ", prog);
		perror(nhistory);
		exit(1);
		/* NOTREACHED */
	    }
	}

#ifdef EXPIRES_DBZ
	fexp = fopen(expires, "a+");
	if ((fexp == NULL) || (dbminit(expires) != 0)) {
	        (void) fprintf(stderr, "%s: open expires db: ", prog);
		perror(expires);
		if (fexp)
		      (void) fclose(fexp);
		fexp = NULL;
	} else {
	        dbzincore(1);
	}
#else
	/* If requested, create the explicit expires file */
	fexp = 0;
	if (doexpires && (fexp = fopen(nexpires, "w")) == 0) {
		(void) fprintf(stderr, "%s: fopen(): ", prog);
		perror(nexpires);
		/* Non-fatal error */
	}
#endif

	/* Process the history file */
	status = histtrim(fin, fout, fexp);

	/* Check for errors and close the files */
	if (ferror(fin)) {
		(void) fprintf(stderr, "%s: ferror(): ", prog);
		perror(history);
		status |= 1;
	}
	if (ferror(fout)) {
		(void) fprintf(stderr, "%s: ferror(): ", prog);
		perror(nhistory);
		status |= 1;
	}
	if (fexp && ferror(fexp)) {
		(void) fprintf(stderr, "%s: ferror(): ", prog);
		perror(nexpires);
		/* Non-fatal error */
		(void)fclose(fexp);
		fexp = 0;
	}
	(void)fclose(fin);
	(void)fclose(fout);
	if (fexp) {
#ifdef EXPIRES_DBZ
	        dbmclose();
#endif
		(void)fclose(fexp);
	}

#ifndef EXPIRES_DBZ
	/* Rename the explicit expires file */
	if (fexp && rename(nexpires, expires) < 0 &&
	    unlink(expires) < 0 &&
	    rename(nexpires, expires) < 0) {
		(void)fprintf(stderr, "%s: rename(): %s -> ", prog, nexpires);
		perror(expires);
		/* Non-fatal error */
	}
#endif

	if (verbose) {
		(void) fflush(stderr);
		(void) printf("%s: %d bad\n", prog, bad);
		(void) printf("%s: %d newly created \"no groups\"\n",
		    prog, newnogroup);
		(void) printf("%s: %d old \"no groups\"\n", prog, oldnogroup);
		(void) printf("%s: %d \"no groups\" omitted\n", prog, nogroup);
		(void) printf("%s: %d not in the feedback file\n",
		    prog, nofeedback);
		(void) printf("%s: %d neverexpired\n", prog, neverexpired);
		(void) printf("%s: %d numread\n", prog, numread);
		(void) printf("%s: %d emitted\n", prog, emitted);
		(void) printf("%s: Took %s\n",
		    prog, fmtdelta(time(0) - currenttime));
	}

	exit(status);
}

int
histtrim(fin, fout, fexp)
	FILE *fin, *fout, *fexp;
{
	/* XXX want this to be a register */
	char *cp, *cp2;
	register int h;
	register int numkept;
	time_t t, ht;
	char line[4096];
	register int n;
	register int hasexpires;
	int keep;
#ifdef EXPIRES_DBZ
	datum dbkey, dbval;
	char buf[1024];
	int ofs;
	time_t et;
#endif

	n = 1;

	while (fgets(line, sizeof(line), fin)) {
		/* Statistics */
		++numread;

		/* Flag the fact that we haven't counted these yet */
		numkept = -1;

		/* Assume no explicit expires date */
		hasexpires = 0;

		/* Check & fix syntax */
		cp = line + strlen(line) - 1;
		if (*cp != '\n') {
		       (void) fprintf(stderr, "%s: parse error #0 line %d\n",
			    prog, n);
		       ++bad;
		       goto emit;
		}
		--cp;
		while ((cp>line) && (isspace(*cp)))
		       --cp;
		*++cp = '\n'; *++cp = '\0';

		/* Step over message id */
		cp = index(line, '\t');
		if (cp == 0) {
			(void) fprintf(stderr, "%s: parse error #1 line %d\n",
			    prog, n);
			++bad;
			goto emit;
		}
		++cp;

		/* Arrival timestamp */
		t = atol(cp);

		/* Optionally look for explicit expires */
		if (fexp) {
			cp = index(cp, '~');
			if (cp == 0) {
				(void) fprintf(stderr,
				    "%s: parse error #2 line %d\n", prog, n);
				++bad;
				goto emit;
			}

			/* Step over '~' */
			++cp;

			/* Look for explicit expires time and date */
			hasexpires = (*cp != '-');
#ifdef EXPIRES_DBZ			
			if (hasexpires) {
			        /* parse expires time */
			        strdcpy(buf, cp, '~');
				et = atoi(buf);
				if (!et)
					et = get_date(buf, NULL);
				if ((et < 0) || (et > maxtime)) {
				        hasexpires = 0;
					/* kill the expiry date */
					cp2 = cp + strlen(buf);
				        *cp++ = '-';
					if (cp2 && (cp2!=cp))
					        bcopy(cp2, cp, strlen(cp2)+1);
					if (et<0) {
					    (void)fprintf(stderr,
						  "%s: bad expiry date '%s'",
						  prog, buf);
					    strdcpy(buf, line, '\0');
					    (void)fprintf(stderr,
							  " in %s\n", buf);
					}
				}
			}
#endif
		}

		/* Find newsgroups */
		cp = index(cp, '\t');

		/* Check for newsgroups */
		if (cp == 0) {
			/* No newsgroups */
			if (t < mintime) {
				/* History entry is too old; suppress */
				++oldnogroup;
				continue;
			}
			/* Keep history entry */
			++nogroup;
			goto emit;
		}

		/* Point to start of newsgroups */
		++cp;

		/* Save pointer to start of first newsgroup */
		cp2 = cp;

		/* Loop through newsgroup(s) */
		keep = numkept = 0;
		for (;;) {
			/* Hash groupname and get group time */
			h = hashgroup(&cp2);
			ht = hashfind(h, cp);
			/* Keep all groups if no feedback file */
			if (ht >= 0)
				keep = 0;

			if (ht < 0) {
				/* Not in feedback file */
				++nofeedback;

				/* Keep history entry if article exists */
				if (!keep && artexists(cp))
					++keep;
			} else if (ht == 0) {
				/* "Never expired" newsgroup */
				++neverexpired;

				/* Keep history entry if article exists */
				if (artexists(cp))
					++keep;
			} else if (t >= ht) {
				/* Keep if not too old */
				++keep;
			}

#ifdef EXPIRES_DBZ
			if (hasexpires && (fexp!=NULL)) {
			        strtcpy(buf, cp);
				dbkey.dptr = buf;
				dbkey.dsize = strlen(buf)+1;
				dbval = dbzfetch(dbkey);
				if (keep && (dbval.dptr == NULL)) {
				       /* Register new entry in expires */
				       ofs = ftell(fexp);
				       (void)fprintf(fexp,
					      "%s\t%ld\n", buf, et);
				       (void)fflush(fexp);
				       dbval.dptr = (char *)&ofs;
				       dbval.dsize = sizeof(ofs);
				       if (store(dbkey, dbval)!=0) {
					    (void)fprintf(stderr, "%s: store of %s failed\n", prog, buf);
				       }
			       } else if (dbval.dptr != NULL) {
				       /* Keep if it has expires line */
				       keep = 1;
			       }
			}
#endif
			/* Update number of groups kept */
			numkept += keep;

			/* Step over article number */
			if (*cp2 != '/') {
				(void) fprintf(stderr,
				    "%s: parse error #3 on line %d\n",
				    prog, n);
				++bad;
				goto emit;
			}
			++cp2;
			while (isdigit(*cp2))
				++cp2;

			if (*cp2 == '\n') {
				/* End of line; all done */
				if (!keep) {
					--cp;
					if (!isspace(*cp))
						++cp;
					*cp++ = '\n';
					*cp = '\0';
				}
				break;
			} else if (*cp2 == ' ') {
				/* Step over whitespace */
				++cp2;
				if (!keep) {
					/* Remove newsgroup */
					bcopy(cp2, cp, strlen(cp2) + 1);
					cp2 = cp;
				} else
					cp = cp2;
			} else {
				(void) fprintf(stderr,
				    "%s: parse error #4 on line %d\n",
				    prog, n);
				goto emit;
			}
		}

		if (numkept == 0) {
			/* History entry now has "no groups" */
			++newnogroup;

			/* Suppress if too old */
			if (t < mintime) {
				++oldnogroup;
				continue;
			}
			++nogroup;
		}
emit:
		/* Output this (possibly modified) history line */
		fputs(line, fout);
		++n;

#ifndef EXPIRES_DBZ
		/* Only add to expires file if there are newsgroups */
		if (hasexpires && numkept != 0) {
			/* Look for at least one newsgroup if necessary */
			if (numkept < 0 &&
			    (cp = index(line, '\t')) && index(cp, '\t') == 0)
				numkept = 1;
			if (numkept > 0)
				fputs(line, fexp);
		}
#endif
	}
	emitted = n - 1;
	return(0);
}

/* Check if an article actually exists */
static int
artexists(art)
	register char *art;
{
	register int i;
	register char *path;


	path = artpath(spool_dir, art);
	i = access(path, F_OK);
	if (debug > 1)
		(void) fprintf(stderr, "%s: checking (%s): %s\n",
		    prog, i < 0 ? "fails" : "succeeds", path);
	if (i < 0)
		return(0);
	return(1);
}

/* Read and parse the feedback file */
int
readfeedback(f, fn)
	register FILE *f;
	int (*fn)();
{
	register int n;
	register char *cp, *cp2;
	char line[1024];
	time_t t;

	n = 0;
	while (fgets(line, sizeof(line), f)) {
		++n;
		cp = line;
		/* String trailing newline */
		cp2 = cp + strlen(cp) - 1;
		if (cp2 >= cp && *cp2 == '\n')
			*cp2++ = '\0';
		t = atoi(cp);
		if ((cp = index(cp, '\t')) == 0) {
			(void) fprintf(stderr,
			    "%s: feedback file syntax error #1 line %d\n",
			    prog, n);
			return(0);
		}
		++cp;
		/* Convert '/'s to '.'s a la the history file */
		for (cp2 = cp; *cp2 != '\0'; ++cp2)
			if (*cp2 == '/')
				*cp2 = '.';
		(void)(*fn)(cp, t);
	}

	/* Paranoid safety check */
	if (ferror(f)) {
		(void) fprintf(stderr, "%s: ", prog);
		perror("readfeedback()");
		return(0);
	}

	return(1);
}
